<?php

namespace app\modules\page\backend;

use app\backend\actions\PropertyHandler;
use app\backend\events\BackendEntityEditEvent;
use app\models\Object;
use app\models\Property;
use app\modules\image\models\Image;
use app\modules\image\widgets\views\AddImageAction;
use app\modules\page\models\Page;
use app\modules\page\models\PageCategory;
use app\models\ViewObject;
use app\modules\page\models\PageImages;
use app\properties\HasProperties;
use app\modules\image\widgets\RemoveAction;
use app\modules\image\widgets\SaveInfoAction;
use app\modules\image\widgets\UploadAction;
use devgroup\JsTreeWidget\AdjacencyFullTreeDataAction;
use devgroup\JsTreeWidget\TreeNodeMoveAction;
use devgroup\JsTreeWidget\TreeNodesReorderAction;
use Yii;
use yii\db\Query;
use yii\web\Response;
use yii\helpers\Html;
use yii\web\BadRequestHttpException;
use yii\filters\AccessControl;
use yii\helpers\ArrayHelper;
use yii\helpers\Url;
use yii\web\NotFoundHttpException;
use yii\web\UploadedFile;
use app\components\Helper; 
use kartik\widgets\ActiveForm;
use arogachev\excel\import\basic\Importer;

class PageController extends \app\backend\components\BackendController
{
    const BACKEND_PAGE_EDIT = 'backend-page-edit';
    const BACKEND_PAGE_EDIT_SAVE = 'backend-page-edit-save';
    const BACKEND_PAGE_EDIT_FORM = 'backend-page-edit-form';

    public function behaviors()
    {
        return [
            'access' => [
                'class' => AccessControl::className(),
                'rules' => [
                    [
                        'allow' => true,
                        'roles' => ['content manage'],
                    ],
                ],
            ],
        ];
    }

    public function actions()
    {
        return [
            'getTree' => [
                'class' => AdjacencyFullTreeDataAction::className(),
                'class_name' => Page::className(),
                'model_label_attribute' => 'name',
            ],
            'addImage' => [
                'class' => AddImageAction::className(),
            ],
            /*
            'upload' => [
                'class' => UploadAction::className(),
                'upload' => 'theme/resources/product-images',
            ],
            */
            'remove' => [
                'class' => RemoveAction::className(),
                'uploadDir' => 'theme/resources/product-images',
            ],
            'save-info' => [
                'class' => SaveInfoAction::className(),
            ],
            'property-handler' => [
                'class' => PropertyHandler::className(),
                'modelName' => Page::className()
            ],
            'move' => [
                'class' => TreeNodeMoveAction::className(),
                'className' => Page::className(),
                'saveAttributes' => ['slug_compiled'],
            ],
            'reorder' => [
                'class' => TreeNodesReorderAction::className(),
                'className' => Page::className(),
            ],
        ];
    }

    public function actionIndex($parent_id = 1)
    {
        $searchModel = new Page();
        $searchModel->parent_id = $parent_id;

        $params = Yii::$app->request->get();
        $dataProvider = $searchModel->search($params);

        $model = null;
        if ($parent_id > 0) {
            $model = Page::find()->where(['id'=>$parent_id])->one();
        }

        return $this->render(
            'index',
            [
                'dataProvider' => $dataProvider,
                'searchModel' => $searchModel,
                'model' => $model,
            ]
        );
    }
        
    public function actionEdit($parent_id, $id = null)
    {
        $model = new Page;
        $model->published = 1;
        if ($id !== null) {
            $model = Page::find()->where(['id'=>$id])->multilingual()->one();
            if ($model === null) {
                throw new NotFoundHttpException;
            }
        }
        $model->parent_id = $parent_id;

        $post = \Yii::$app->request->post();

        if ($model->load($post)) {

            if ($model->validate()) {

                $save_result = $model->save();
                $model->saveProperties($post);

                if ($save_result) {

                    $this->updateMainImage($model);

                    $this->updateSortImage($model);

                    $this->runAction('save-info', ['model_id' => $model->id]);
                    Yii::$app->session->setFlash('info', Yii::t('app', 'Object saved'));
                    $returnUrl = Yii::$app->request->get('returnUrl', ['/page/backend/index']);
                    switch (Yii::$app->request->post('action', 'save')) {
                        case 'next':
                            return $this->redirect(
                                [
                                    '/page/backend/edit',
                                    'returnUrl' => $returnUrl,
                                    'parent_id' => Yii::$app->request->get('parent_id', null)
                                ]
                            );
                        case 'back':
                            return $this->redirect($returnUrl);
                        default:
                            return $this->redirect(
                                Url::toRoute(
                                    [
                                        '/page/backend/edit',
                                        'id' => $model->id,
                                        'returnUrl' => $returnUrl,
                                        'parent_id' => $model->parent_id
                                    ]
                                )
                            );
                    }
                } else {
                    Yii::$app->session->setFlash('error', Yii::t('app', 'Cannot update data'));
                }
            }
        }

        return $this->render(
            'edit',
            [
                'model' => $model,
            ]
        );
    }

    public function actionUpdatePage()
    {
        Yii::$app->response->format = Response::FORMAT_JSON;
        $post = Yii::$app->request->post();
       
        if (!isset($post['editableIndex'], $post['editableKey'])) {
            throw new BadRequestHttpException;
        }
        
        $id = $post['editableKey'];
        /** @var User Package */
        $model = Page::findOne($id);
        if ($model === null) {
            throw new BadRequestHttpException;
        }
        
        $index = $post['editableIndex'];
        if (!is_array($post['Page'][$index])) {
            throw new BadRequestHttpException;
        }
        
        if (count($post['Page'][$index])===0) {
            throw new BadRequestHttpException;
        }
        
        $modifiedAttribute = array_keys($post['Page'][$index])[0];
       
        $model->setAttributes($post['Page'][$index]);
        
         
        if (!$model->save()) {
            return [
                'message' => Yii::t('app', 'Cannot save object'),
            ];
        }
       
        if ($modifiedAttribute === 'published') {
            
            if ($model->published == 0) {
                    $label_class = 'label-default';
                    $value = 'Inactive';
                }elseif ($model->published == 1) {
                    $label_class = 'label-success';
                    $value = 'Active';}
                
                $model->published = Html::tag(
                    'span',
                    Yii::t('app', $value),
                    ['class' => "label $label_class"]
                );
        }
        
        return [
            'output' => $model->getAttribute($modifiedAttribute),
        ];
    }

    public function actionDelete($id = null)
    {
        if ((null === $id) || (null === $model = Page::find()->where(['id'=>$id])->multilingual()->one())) {
            throw new NotFoundHttpException;
        }

        if($model->delete())
            $delete = true;
        else
           $delete = false;
        
        if (!$delete) {
            Yii::$app->session->setFlash('info', Yii::t('app', 'The object is placed in the cart'));
        } else {
            Yii::$app->session->setFlash('info', Yii::t('app', 'Object removed'));
        }

        return $this->redirect(
            Yii::$app->request->get(
                'returnUrl',
                Url::toRoute(['index', 'parent_id' => $model->parent_id])
            )
        );
    }

    public function actionRemoveAll($parent_id)
    {
        $items = Yii::$app->request->post('items', []);
        if (!empty($items)) {
            $items = Page::find()->where(['in', 'id', $items])->multilingual()->all();
            foreach ($items as $item) {
                $this->deleteFile($item->image);
                $item->delete();
            }
        }

        return $this->redirect(['index', 'parent_id' => $parent_id]);
    }

    public function actionRestore($id = null, $parent_id = null)
    {
        if (null === $id) {
            new NotFoundHttpException();
        }

        if (null === $model = Page::find()->where(['id'=>$id])->multilingual()->one()) {
            new NotFoundHttpException();
        }

        $model->restoreFromTrash();

        Yii::$app->session->setFlash('success', Yii::t('app', 'Object successfully restored'));

        return $this->redirect(
            Yii::$app->request->get(
                'returnUrl',
                Url::toRoute(['edit', 'id' => $id, 'parent_id' => $parent_id])
            )
        );
    }

    public function actionUpload($id) {

        Yii::$app->response->format = Response::FORMAT_JSON;
        $model                = Page::find()->where(['id'=>$id])->one();
        $pageImage            = new PageImages;
        $pageImage->file_name =  $pageImage->UploadFile($model);
        $pageImage->object_id = $model->id;
        $pageImage->main      = 0;

        if($pageImage->save())
            return true;
        else
            return false;

    }

    public function updateMainImage(Page $model)
    {
        if(Yii::$app->request->post('mainImageId'))
        {
            // Clear current main image
            PageImages::updateAll(['main'=>0],'object_id ='.$model->id);
            // Set new main image
            $image = PageImages::find()->where('id = '.Yii::$app->request->post('mainImageId').' AND object_id ='.$model->id)->one();
            $image->main = 1;
            $image->update();
        }
    }

    public function updateSortImage(Page $model)
    {

        $soerImage = Yii::$app->request->post('sortImageId');

        $countImages = sizeof($soerImage);

        for($i = 0 ; $i < $countImages; $i++) {
            $image = PageImages::find()->where('id = '.$soerImage[$i].' AND object_id ='.$model->id)->one();
            $image->sort = $i;
            $image->update();
        }

        if(Yii::$app->request->post('sortImageId'))
        {
            // Clear current main image
            PageImages::updateAll(['main'=>0],'object_id ='.$model->id);
            // Set new main image
            $image = PageImages::find()->where('id = '.Yii::$app->request->post('mainImageId').' AND object_id ='.$model->id)->one();
            $image->main = 1;
            $image->update();
        }
    }

    public function actionAjaxDeleteImage(){

        if(Yii::$app->request->post('key')) {

            $id = Yii::$app->request->post('key');

            $image = PageImages::find()->where('id = '.$id)->one();

            if($image AND $image->delete())
                return true;
            else
                throw new NotFoundHttpException('Error delete file !');
        }
    }

    public function UloadFileImport($model,$arrt)
    {
        $image = UploadedFile::getInstance($model,$arrt);
        if ($image !== null)
        {

            if (file_exists(Yii::getAlias('@app/web/upload/import/encyclopedia.xlsx')))
                unlink(Yii::getAlias('@app/web/upload/import/encyclopedia.xlsx'));

            $prefx = time();
            $path = '/upload/import/encyclopedia.'.strtolower(substr(strrchr($image->name,"."),1));
            $stream = fopen($image->tempName, 'r+');
            Yii::$app->getModule('page')->fsComponent->writeStream($path, $stream);
            fclose($stream);

            return $path;
        }else
            return false;
    }

    public function actionImport()
    {
        $request = Yii::$app->request;
        $pageCategory = PageCategory::find()->where(['slug'=>'encyclopedia'])->one();
        $modelPage = new Page;
        $highestRow = 0;

        if ($modelPage->load($request->post()) OR $request->get('file')) {


            if($request->get('file') AND file_exists($request->get('file'))){
                $file = $request->get('file');
            }else{

                if ($import_excel = $modelPage->UploadImportFile($modelPage)) {
                    $modelPage::deleteAll('category_id = ' . $pageCategory->id);
                    $file = $import_excel;// Yii::getAlias('@app/web/upload/import/encyclopedia.xlsx');
                }
            }
            try {

                $fileType = \PHPExcel_IOFactory::identify($file);
                $objRrader = \PHPExcel_IOFactory::createReader($fileType);
                $objPHPExcel = $objRrader->load($file);
            } catch (Exception $e) {

                die("Error");
            }

            $sheet = $objPHPExcel->getSheet(0);
            $highestRow = $sheet->getHighestRow();
            $highestColumn = $sheet->getHighestColumn();
        }

            $i = 1;

           if($request->get('limit'))
                $row = $request->get('limit');
            else
                $row = 1;

            while($row <= $highestRow){

                if($i == 50){
                    $this->redirect(['/page/backend/import','limit'=>$row,'file'=>$file]);
                    break;
                 }

                $page = new Page;
                $array_excel = $sheet->rangeToArray("A".$row.':'.$highestColumn.$row,NULL,true,true)[0];


                $slug = Helper::createSlug(trim(strip_tags($array_excel[0])));
                $model_attr ['slug']                 = $slug;
                $model_attr ['name']                 = trim(strip_tags(($array_excel[0])?$array_excel[0]:$array_excel[0]));
                $model_attr ['title']                = trim(strip_tags(($array_excel[0])?$array_excel[0]:$array_excel[0]));
                $model_attr ['category_id']          = $pageCategory->id;
                $model_attr ['main']                 = 0;
                $model_attr ['content']              = ($array_excel[1])?$array_excel[1]:$array_excel[1];
                $model_attr ['h1']                   = trim(strip_tags(($array_excel[0])?$array_excel[0]:$array_excel[0]));
                $model_attr ['meta_description']     = trim(strip_tags(($array_excel[0])?$array_excel[0]:$array_excel[0]));
                $model_attr ['breadcrumbs_label']    = trim(strip_tags(($array_excel[0])?$array_excel[0]:$array_excel[0]));
                $model_attr ['announce']             = substr(strip_tags(($array_excel[1])?$array_excel[1]:$array_excel[1]),0,250);
                $model_attr ['slug_compiled']        = $slug;
                $model_attr ['published']            = 1;
                $model_attr ['show_type']            = 'show';

                $model_attr ['name_uz']              = trim(strip_tags($array_excel[2]));
                $model_attr ['title_uz']             = trim(strip_tags($array_excel[2]));
                $model_attr ['h1_uz']                = trim(strip_tags($array_excel[2]));
                $model_attr ['meta_description_uz']  = trim(strip_tags($array_excel[2]));
                $model_attr ['breadcrumbs_label_uz'] = trim(strip_tags($array_excel[2]));
                $model_attr ['announce_uz']          = substr(strip_tags($array_excel[3]),0,250);
                $model_attr ['content_uz']           = $array_excel[3];
                $model_attr ['title_uz']             = trim(strip_tags($array_excel[2]));

                $page->parent_id = 1;
                $page->attributes = $model_attr;

                if($page->validate())
                   $page->save();

                $row++;
                $i++;
            }

        return $this->render('import',
            [
                'model' => $modelPage,
            ]);
    }
}
