<?php

namespace app\modules\user\controllers;

use app\backend\components\BackendController;
use app\modules\user\models\User;
use Yii;
use yii\filters\AccessControl;
use yii\filters\VerbFilter;
use yii\helpers\ArrayHelper;
use yii\helpers\Json;
use yii\rbac\Item;
use yii\helpers\Url;
use yii\db\Query;
use yii\web\NotFoundHttpException;
use app\backend\actions\UpdateEditable;
use kartik\helpers\Html;
use yii\web\Response;
use yii\web\BadRequestHttpException;

/**
 * BackendUserController implements the CRUD actions for User model.
 */
class BackendUserController extends BackendController
{
    /**
     * @inheritdoc
     */
    public function behaviors()
    {
        return [
            'access' => [
                'class' => AccessControl::className(),
                'rules' => [
                    [
                        'allow' => true,
                        'roles' => ['user manage'],
                    ],
                ],
            ],
            'verbs' => [
                'class' => VerbFilter::className(),
                'actions' => [
                    'delete' => ['post'],
                ],
            ],
        ];
    }
    
    
    public function actions()
    {
        return [
            'update-editable' => [
                'class' => UpdateEditable::className(),
                'modelName' => User::className(),
                'scenario'=>'status',
                'allowedAttributes' => [
                    'status' => function (user $model) {
                        
                       if ($model === null || $model->status === null) {
                                return null;
                            }
                             if ($model->status == 0) {
                                $label_class = 'label-default';
                                $value = Yii::t('app', 'Deleted');
                            }elseif ($model->status == 10) {
                                $label_class = 'label-success';
                                $value = Yii::t('app', 'Active');
                            }elseif ($model->status == 5) {
                                $label_class = 'label-danger';
                                $value = Yii::t('app', 'Удалить');
                            }
                                   
                            return Html::tag(
                                'span',
                                Yii::t('app', $value),
                                ['class' => "label $label_class"]
                            );
                    },
                ],
            ],
        ];
    }

    public function actionAjaxListData($q = null, $id = null) {

        Yii::$app->response->format = \yii\web\Response::FORMAT_JSON;
        $out = ['results' => ['id' => '', 'text' => '']];
        if (!is_null($q)) {
            $query = new Query;
            $query->select('u.id, u.username, phone, ut.first_name, ut.last_name')
                ->from('user u')
                ->leftJoin('userTranslate ut','ut.user_id = u.id' )
                ->groupBy('u.id')
                ->where('username LIKE "%' . $q .'%"')
                ->orWhere('ut.first_name LIKE "%' . $q .'%"')
                ->orWhere('ut.last_name LIKE "%' . $q .'%"')
                ->orWhere(['phone'=>$q])
                ->limit(100);
            $command = $query->createCommand();
            $data = $command->queryAll();
            $out['results'] = array_values($data);
        }
        /*
        elseif ($id > 0) {
            $out['results'] = ['id' => $id, 'text' => City::find($id)->name];
        }
        */
        return $out;
    }

    /**
     * Lists all User models.
     * @return mixed
     */
    public function actionIndex()
    {

        $searchModel = new User();
        $searchModel->scenario = 'search';

        $params = Yii::$app->request->get();
        $dataProvider = $searchModel->search($params);

        return $this->render(
            'index',
            [
                'dataProvider' => $dataProvider,
                'searchModel' => $searchModel,
            ]
        );
    }

    public function deleteFile($file)
    {
        if(trim($file) != "")
        {
          if(file_exists(Yii::$app->basePath."/web/files/".$file))
            {
                 Yii::$app->getModule('user')->fsComponent->delete($file);
                return true;
            }else
                return false;
        }
    }

    public function actionUpdate($id = null)
    {
        
        /** @var User $model */
        if (is_null($id)) {
            $model = new User(['scenario' => 'adminSignup']);
            $model->generateAuthKey();
        } else {
            $model = $this->findModel($id);
            $model->scenario = 'admin';
        }
        
        $assignments = Yii::$app->authManager->getAssignments($id);
        
        if (Yii::$app->request->isPost) {

            $model->load(Yii::$app->request->post());
            $model->auth_key = '';
            if ($model->validate()) {

                if ($id !== null && !empty($model->password)) {
                    $model->setPassword($model->password);
                }

                if($image = $model->uploadFile()){
                    $model->image = $image;
                    $model->deleteImage();
                }
                $model->save();
                $postAssignments = Yii::$app->request->post('AuthAssignment', []);
                $errors = [];
                foreach ($assignments as $assignment) {
                    $key = array_search($assignment->roleName, $postAssignments);
                    if ($key === false) {
                        Yii::$app->authManager->revoke(new Item(['name' => $assignment->roleName]), $model->id);
                    } else {
                        unset($postAssignments[$key]);
                    }
                }
                foreach ($postAssignments as $assignment) {
                    try {
                        Yii::$app->authManager->assign(new Item(['name' => $assignment]), $model->id);
                    } catch (\Exception $e) {
                        $errors[] = 'Cannot assign "' . $assignment . '" to user';
                    }
                }
                if (count($errors) > 0) {
                    Yii::$app->getSession()->setFlash('error', implode('<br />', $errors));
                }
                Yii::$app->session->setFlash('success', Yii::t('app', 'Record has been saved'));
                $returnUrl = Yii::$app->request->get('returnUrl', ['index']);
                switch (Yii::$app->request->post('action', 'save')) {
                    case 'next':
                        return $this->redirect(
                            [
                                'update',
                                'returnUrl' => $returnUrl,
                            ]
                        );
                    case 'back':
                        return $this->redirect($returnUrl);
                    default:
                        return $this->redirect(
                            Url::toRoute(
                                [
                                    'update',
                                    'id' => $model->id,
                                    'returnUrl' => $returnUrl
                                ]
                            )
                        );
                }
            }
        }
        return $this->render(
            'update',
            [
                'model' => $model,
                'assignments' => ArrayHelper::map($assignments, 'roleName', 'roleName'),
            ]
        );
    }
    
    public function actionUpdateUser()
    {
        Yii::$app->response->format = Response::FORMAT_JSON;
        $post = Yii::$app->request->post();
       
        if (!isset($post['editableIndex'], $post['editableKey'])) {
            throw new BadRequestHttpException;
        }
        
        $id = $post['editableKey'];
        /** @var User Package */
        $model = User::findOne($id);
        if ($model === null) {
            throw new BadRequestHttpException;
        }
        
        $index = $post['editableIndex'];
        if (!is_array($post['User'][$index])) {
            throw new BadRequestHttpException;
        }
        
        if (count($post['User'][$index])===0) {
            throw new BadRequestHttpException;
        }
        
        $modifiedAttribute = array_keys($post['User'][$index])[0];

        $model->setAttributes($post['User'][$index]);
       
         if ($modifiedAttribute === 'free_message') {
             $model->scenario = 'free_message';
             $model->free_message = $post['User'][$index]['free_message'] ;
             $model->save();
        }
         if ($modifiedAttribute === 'sort') {
             $model->scenario = 'sort';
             $model->sort = $post['User'][$index]['sort'] ;
             $model->save();
        }
        
        return [
            'output' => $model->getAttribute($modifiedAttribute),
        ];
    }

    /**
    * Deletes an existing User model.
    * If deletion is successful, the browser will be redirected to the 'index' page.
    * @param integer $id
    * @return mixed
    */
    public function actionDelete($id)
    {
        $model = $this->findModel($id);
        $model->scenario = 'admin';
        $model->status = User::STATUS_DELETED;
        $this->deleteFile($model->image);
        $model->delete();
        //$model->save(true, ['status']);
        return $this->redirect(['index']);
    }

    public function actionRemoveAll()
    {
        $items = Yii::$app->request->post('items', []);
        if (!empty($items)) {
            $items = User::find()->where(['in', 'id', $items])->all();
            foreach ($items as $item) {
                $item->scenario = 'admin';
                $item->status = User::STATUS_DELETED;
                $item->save(true, ['status']);
            }
        }

        return $this->redirect(['index']);
    }

    public function actionAddAssignment($id, $userId)
    {
        try {
            Yii::$app->authManager->assign(new Item(['name' => $id]), $userId);
            $result = [
            'status' => 1,
            'message' => 'Success',
            ];
        } catch (\Exception $e) {
            $result = [
                'status' => 0,
                'message' => 'Cannot add assignment',
            ];
        }
        return Json::encode($result);
    }

    public function actionRemoveAssignment($id, $userId)
    {
        if (Yii::$app->authManager->revoke(new Item(['name' => $id]), $userId)) {
            $result = [
                'status' => 1,
                'message' => 'Success',
            ];
        } else {
            $result = [
            'status' => 0,
            'message' => 'Cannot remove assignment',
            ];
        }
        return Json::encode($result);
    }
    
    public function afterDelete()
    {
        parent::afterDelete();
    }


    /**
    * Finds the User model based on its primary key value.
    * If the model is not found, a 404 HTTP exception will be thrown.
    * @param integer $id
    * @return User the loaded model
    * @throws NotFoundHttpException if the model cannot be found
    */
    protected function findModel($id)
    {
        if (($model = User::find()->where(['id'=>$id])->multilingual()->one()) !== null) {
            return $model;
        } else {
            throw new NotFoundHttpException('The requested page does not exist.');
        }
    }
}
