<?php

namespace app\modules\user\models;

use Yii;
use app\modules\review\models\Rating;
use app\modules\review\models\Review;
use app\modules\regions\models\City;
use app\modules\regions\models\Country;
use yii\base\NotSupportedException;
use yii\behaviors\TimestampBehavior;
use yii\caching\TagDependency;
use yii\db\ActiveRecord;
use yii\web\IdentityInterface;
use yii\data\ActiveDataProvider;
use yii\helpers\ArrayHelper;

use devgroup\TagDependencyHelper\ActiveRecordHelper;

use app\modules\core\models\Language;
use omgdef\multilingual\MultilingualBehavior;

use yii\imagine\Image;
use Imagine\Image\ManipulatorInterface;

use app\modules\user\models\UserRating;
Use app\modules\api\models\ApplePushNotification;
use app\modules\api\models\GoogleCloudMessaging;

use yii\web\UploadedFile;

use app\modules\notification\models\Notification;
/**
 * Class User
 * @package app\models
 *
 * @property integer $id
 * @property string $username
 * @property string $password_hash
 * @property string $password_reset_token
 * @property string $email
 * @property string $auth_key
 * @property integer $status
 * @property integer $create_time
 * @property integer $update_time
 * @property string $first_name
 * @property string $last_name
 * @property string $middle_name
 * @property \app\modules\user\models\UserService[] $services
 * @property string $displayName  Display name for the user visual identification
 * Relations:
 */
class User extends ActiveRecord implements IdentityInterface
{
    const STATUS_DELETED = 0;
    const STATUS_ACTIVE = 10;
    const EMAIL_ACTIVE = 1;
    const USER_GUEST = 0;
    const DELETED = 5;

    const TOKEN_ANDROID = 'android';
    const TOKEN_IOS     = 'ios';

    public $confirmPassword;
    public $confirmPassword_signup;
    public $newPassword;
    public $password;

    public $uploadDir = '/upload/user/';
    public $thumbnailDir = '/thumbnail/user';

    public function rules()
    {
        return [
            //status
            ['status', 'default', 'value' => self::STATUS_ACTIVE],
            ['status', 'in', 'range' => [self::STATUS_ACTIVE, self::STATUS_DELETED,self::DELETED]],

            //username
            ['username', 'filter', 'filter' => 'trim'],
            ['username', 'required'],
            ['username', 'string', 'min' => 2, 'max' => 255],
            ['username', 'unique'],

            //seo
            [['title', 'keywords', 'description','key'], 'string', 'max' => 255],

            // email
            ['email', 'filter', 'filter' => 'trim'],
            ['email', 'email'],
            ['email', 'unique', 'message' => Yii::t('app', 'This email address has already been taken.')],
            ['email', 'exist', 'message' => Yii::t('app', 'There is no user with such email.'), 'on' => 'requestPasswordResetToken'],
            ['email', 'required', 'on' => 'admin-email'],

            // phone
            ['phone', 'unique'],
            ['phone', 'required'],
            ['phone', 'match', 'pattern' => "/^[0-9]{12}$/",'message'=> Yii::t('app', 'Телефон должен быть вида 998977131212')],

            [['first_name', 'last_name', 'middle_name'], 'string', 'max' => 255],
            [['first_name','last_name'], 'required'],

            // password
            ['password', 'string', 'min' => 8],
            ['password_code', 'required'],
            ['password_code', 'number'],

            // change password
            [['newPassword', 'confirmPassword'], 'required'],
            [['newPassword', 'confirmPassword'], 'string', 'min' => 8],
            [['confirmPassword'], 'compare', 'compareAttribute' => 'newPassword'],

            // for signup
            [['confirmPassword_signup'], 'required'],
            ['password', 'required','on'=>['signup']],
            ['confirmPassword_signup', 'compare', 'compareAttribute' => 'password'],

            // image
            ['image', 'file', 'extensions' => ['png', 'jpg'],'maxSize' => 1024 * 1024 * 2 ],
            ['image', 'isFileExtensions'],

            [['city_id','country_id'], 'integer'],
            [['rating','last_login','gender','facebook','twitter','instagram'],'safe'],
        ];
    }

    public function scenarios()
    {
        return [
            'signup'                    => ['username', 'password', 'confirmPassword_signup', 'first_name', 'last_name'],
            'resetPassword'             => ['password_hash','password_reset_token'],
            'password_hash'             => ['password_hash','password_code','password_reset_date'],
            'password_code'             => ['password_hash'],
            'requestPasswordResetToken' => ['email'],
            'status'                    => ['status'],
            'active'                    => ['active'],
            'sort'                      => ['sort'],
            'rating'                    => ['rating'],
            'last_login'                => ['last_login'],
            'registerService'           => ['email', 'username', 'first_name', 'last_name'],
             //update profile for user
            'changePassword'            => ['password', 'newPassword', 'confirmPassword'],

            //admin search
            'search'                    => [
                                           'id', 
                                           'username', 
                                           'email', 
                                           'status', 
                                           'create_time', 
                                           'first_name', 
                                           'last_name',
                                           ],
            //for admin update
            'admin'                     => [
                                            'username', 
                                            'password',
                                            'status', 
                                            'email', 
                                            'first_name', 
                                            'last_name',
                                            'middle_name',
                                            'city_id',
                                            'country_id',
                                            'title',
                                            'keywords',
                                            'description',
                                            'facebook',
                                            'twitter',
                                            'instagram'
                                           ],
            'adminSignup'               => ['username', 'status', 'email', 'password', 'first_name', 'last_name', 'auth_key'],
            'passwordResetToken'        => ['password_reset_token'],
            'image'                     => ['image'],
        ];
    }

    public function isFileExtensions()  {
        if ($this->hasErrors() === false) {
           if(!(strtolower(substr(strrchr($this->image,"."),1)) == 'png' OR strtolower(substr(strrchr($this->image,"."),1)) == 'jpg'))
            {
                $this->addError('file',Yii::t('yii', 'Only files with these extensions are allowed: png jpg.'));
            }
        }
    }

    public static function tableName() {
        return '{{%user}}';
    }

    public function attributeLabels() {
        return [
            'id'                     => Yii::t('app', 'ID'),
            'username'               => Yii::t('app', 'Username'),
            'auth_key'               => Yii::t('app', 'Auth Key'),
            'password'               => Yii::t('app', 'Password'),
            'password_reset_token'   => Yii::t('app', 'Password Reset Token'),
            'email'                  => Yii::t('app', 'E-mail'),
            'status'                 => Yii::t('app', 'Status'),
            'create_time'            => Yii::t('app', 'Create Time'),
            'update_time'            => Yii::t('app', 'Update Time'),
            'first_name'             => Yii::t('app', 'First Name'),
            'last_name'              => Yii::t('app', 'Last Name'),
            'middle_name'            => Yii::t('app', 'Middle name'),
            'newPassword'            => Yii::t('app', 'New Password'),
            'confirmPassword'        => Yii::t('app', 'Confirm Password'),
            'confirmPassword_signup' => Yii::t('app', 'Confirm Password'),
            'city_id'                => Yii::t('app', 'City'),
            'country_id'             => Yii::t('app', 'Country'),
            'phone'                  => Yii::t('app', 'Phone'),
            ];
    }

    public static function getStatuses()
    {
        return [
                self::STATUS_ACTIVE => Yii::t('app', 'Active'),
                self::STATUS_DELETED => Yii::t('app', 'Deleted'),
               ];
    }

    public function behaviors() {
        return [
            [
                'class' => MultilingualBehavior::className(),
                'languages' =>Language::getAllLanguages('local'),
              //'languageField' => 'language',
              //'localizedPrefix' => '',
              //'requireTranslations' => false',
                'dynamicLangClass' => false,
                'langClassName' => UserTranslate ::className(), // or namespace/for/a/class/PostLang
                'defaultLanguage' => Language::getDefaultLang()->url,
                'langForeignKey' => 'user_id',
                'attributes' => [
                    'title',
                    'keywords',
                    'description',
                ]
            ],
        
            [
                'class' => TimestampBehavior::className(),
                'createdAtAttribute' => 'create_time',
                'updatedAtAttribute' => 'update_time',
            ],
            [
                'class' => \devgroup\TagDependencyHelper\ActiveRecordHelper::className(),
            ],
        ];
    }

    public function search($params) {

        $query = self::find();

        $dataProvider = new ActiveDataProvider(
            [
                'query' => $query,
                'pagination' => [
                    'pageSize' => 20,
                ],
            ]
        );

        if (!($this->load($params))) {
            return $dataProvider;
        }

        $query->andFilterWhere([self::tableName().'.id' => $this->id]);
        $query->andFilterWhere([self::tableName().'.status' => $this->status]);
        $query->andFilterWhere(['like', 'username', $this->username]);
        $query->andFilterWhere(['like', 'phone', $this->phone]);

        return $dataProvider;
    }

    public static function findIdentityByAccessToken($token, $type = null) {
        // see https://github.com/yiisoft/yii2/issues/2689
        throw new NotSupportedException('"findIdentityByAccessToken" is not implemented.');
    }

    public static function findIdentity($id) {

        if (is_numeric($id)) {
            $model = Yii::$app->cache->get("User:$id");
            if ($model === false) {

                $model = static::find()
                    ->with('services')
                    ->where(['id'=>$id,'status'=>static::STATUS_ACTIVE])
                    ->one();
                if ($model !== null) {
                    Yii::$app->cache->set(
                        "User:$id",
                        $model,
                        3600,
                        new TagDependency([
                            'tags' => [
                                ActiveRecordHelper::getObjectTag($model->className(), $model->id)
                            ]
                        ])
                    );
                }

            }
            return $model;
        } else {
            return null;
        }
    }

    public static function findByUsername($username) {
        $byName = static::findOne(['username' => $username, 'status' => static::STATUS_ACTIVE]);
        if($byName)
           return $byName;
        else
           return static::findOne(['email' => $username, 'status' => static::STATUS_ACTIVE]);
    }

    public static function findByPhone($phone)  {
        $byName = static::findOne(['phone' => $phone, 'status' => static::STATUS_ACTIVE]);
        if($byName)
           return $byName;
        else if( $byName = static::findOne(['username' => $phone, 'status' => static::STATUS_ACTIVE]))
           return $byName;
        else
           false;
    }

    public static function findByPasswordResetToken($token) {
        if (static::isPasswordResetTokenValid($token) === false) {
            return null;
        }

    return static::findOne([
            'password_reset_token' => $token,
            'status' => self::STATUS_ACTIVE,
        ]);
    }

    public static function isPasswordResetTokenValid($token) {
        if (empty($token) === true) {
            return false;
        }
        $expire = Yii::$app->modules['user']->passwordResetTokenExpire;
        $parts = explode('_', $token);
        $timestamp = (int) end($parts);
        return $timestamp + $expire >= time();
    }

    public function getId() {
        return $this->id;
    }

    public function getAuthKey() {
        return $this->auth_key;
    }

    public function validateAuthKey($authKey) {
        return $this->getAuthKey() === $authKey;
    }

    public function validatePassword($password) {
        return Yii::$app->security->validatePassword($password, $this->password_hash);
    }
    
    public function isActiveEmail($user) {
        if($user->active == 1)
            return true;
        else
            return false;    
    }

    public function setPassword($password) {
        $this->password_hash = Yii::$app->security->generatePasswordHash($password);
    }

    public function generateAuthKey() {
        $this->auth_key = Yii::$app->security->generateRandomString();
    }

    public function generatePasswordResetToken() {
        $this->password_reset_token = Yii::$app->security->generateRandomString() . '_' . time();
    }
    
    public function generateEmailActiveKey() {
        $this->email_active = Yii::$app->security->generateRandomString() . '_' . time();
    }

    public function removePasswordResetToken() {
        $this->password_reset_token = null;
    }

    public function afterSave($insert, $changedAttributes) {
        parent::afterSave($insert, $changedAttributes);
        /*
        if($insert)
            $action = UserLog::ACTION_CREATE;
        else
            $action = UserLog::ACTION_EDIT;

        if(Yii::$app->user->id)
            UserLog::addLog($this->attributes,$changedAttributes,$action);
        else
            UserLog::addLog($this->attributes,$changedAttributes,$action,$this->id);
    */
        }

    public function uploadFile() {

        $file =  UploadedFile::getInstancesByName('User');

        if(!$file)
            $file =  UploadedFile::getInstancesByName('file');

        $file = $file[0];

        if ($file) {
            $prefx = rand(100,999).'_'.time();
            $fileName  = $this->id.'_'.$prefx.'.'.$file->extension;
            if(!file_exists(Yii::getAlias('@webroot'.$this->uploadDir)))
                mkdir(Yii::getAlias('@webroot'.$this->uploadDir),0777);

            $uploadDir = Yii::getAlias('@webroot'.$this->uploadDir);
            $fn = $uploadDir . $fileName;
            $file->saveAs($fn);
            return str_replace(Yii::getAlias('@webroot'), '', $uploadDir).$fileName;
        }else
            return false;
    }

    public function gravatar($size = 40) {
        $hash = md5(strtolower(trim($this->email)));
        return $this->getThumbnail($size,$size);
    }

    public function getServices() {
        return $this->hasMany(UserService::className(), ['user_id' => 'id']);
    }

    public function getCity()  {
        return $this->hasOne(City::className(), ['id' => 'city_id']);
    }

    public function getCountry() {
        return $this->hasOne(Country::className(), ['id' => 'country_id']);
    }

    public function getDisplayName() {
        $name = trim($this->first_name . ' ' . $this->last_name);
        return $name ? $name : $this->username;
    }

    public function beforeSave($insert) {

        if (!$insert) {
            // reset a cache tag to get a new parent model below
            TagDependency::invalidate(Yii::$app->cache, [ActiveRecordHelper::getCommonTag(self::className())]);
        }

        if (!parent::beforeSave($insert)) {
            return false;
        }

        if($this->isNewRecord){
            $this->setPassword($this->password);
        }
        return true;
    }

    public static function setAttributesClient($attributes,$client_id) {
        $model = new User(['scenario' => 'registerService']); 
        $model->status = User::STATUS_ACTIVE; 
        $model->active = 1;
        if($client_id == 'facebook')
        {
          $model->username = $attributes['user']['name'];
          $model->email = $attributes['user']['email'];
          
        }
        else if($client_id == 'vkontakte')
        {
          $model->username = $attributes['user']['name'];
          $model->email = $attributes['user']['email'];
        }
        else if($client_id == 'twitter')
        {
          $model->username = $attributes['user']['username'];
          $model->email = $attributes['user']['email'];
        }
        return $model;
    }

    public function getListData()  {
        return ArrayHelper::map(User::find()->orderBy('username')->all(),'id','username');
    }

    public function getTokenAndriod() {
        return $this->hasMany(GoogleCloudMessaging::className(), ['user_id' => 'id'])->all();
    }

    public function getTokenIOS() {
        return $this->hasMany(ApplePushNotification::className(), ['user_id' => 'id'])->all();
    }

    public static function find() {
        $query = new UserQuery(get_called_class());
        $query->localized();
        return $query;
    }

    public function  getThumbnail($width,$height) {
       $image = $width."-".$height."_".strtolower(substr(strrchr($this->image,"/"),1));

       $file = Yii::getAlias('@webroot'.$this->thumbnailDir.'/'.$image);

       if(file_exists($file)) {
            return $this->thumbnailDir.'/'.$image;
        }else {

            if(!file_exists(Yii::getAlias('@webroot'.$this->thumbnailDir)))
                mkdir(Yii::getAlias('@webroot'.$this->thumbnailDir),0777);

            if(trim($this->image != "") AND file_exists(Yii::getAlias('@webroot').$this->image)){

                $thumb =  Image::thumbnail('@webroot'.$this->image, $width, $height,ManipulatorInterface::THUMBNAIL_INSET)
                    ->save(Yii::getAlias('@webroot'.$this->thumbnailDir.'/'.$width."-".$height."_".strtolower(substr(strrchr($this->image,"/"),1))), ['quality' => 100]);
                return $this->thumbnailDir.'/'.$image;

            }else{

                if(!file_exists(Yii::getAlias('@webroot').$this->thumbnailDir.'/'.$width.'-'.$height.'_user-no-image.jpg')){
                    $thumb =  Image::thumbnail('images/user-no-image.jpg', $width, $height,ManipulatorInterface::THUMBNAIL_INSET)
                        ->save(Yii::getAlias('@webroot'.$this->thumbnailDir.'/'.$width."-".$height."_user-no-image.jpg"), ['quality' => 100]);
                    return $this->thumbnailDir.'/'.$width.'-'.$height.'_user-no-image.jpg';
                }else
                    return $this->thumbnailDir.'/'.$width.'-'.$height.'_user-no-image.jpg';

            }
        }
    }

    public function sendSms($text) {

        if($this->sms_notification == 1){
            if(!empty($this->phone) AND preg_match("/^998[0-9]{9}$/", $this->phone)) {
                if(Yii::$app->sms->send([$this->phone],$text))
                    return true;
                else
                    return false;
            }else
                return false;
        }else
            return false;
    }

    public function sendPushNotification($message) {

        if($this->push_notification == 1){
            $token_android = false;
            $token_ios     = false;

            foreach ($this->getTokenAndriod() as $item) {
                $token_android[] = $item->gcm_regid;
            }

            foreach ($this->getTokenIOS() as $item) {
                $token_ios[] = $item->deviceToken;
            }

            if($token_android) {
                $gcm = Yii::$app->gcm;
                $result =   $gcm->sendMulti(
                    $token_android,
                    $message,
                    ['customerProperty' => 1],
                    ['timeToLive' => 3]);
            }

            if($token_ios) {
                $apns   = Yii::$app->apns;
                $result = $apns->sendMulti(
                    $token_ios,
                    $message['title'],
                    $message,
                    [
                        'sound' => 'default',
                        'badge' => 1
                    ]);
            }
        }
    }

    public function getPushMessage($text,$options){

        $url = Yii::$app->request->hostInfo;


        // if notification type == question
        if($options['type'] == Notification::TYPE_QUESTION){

            // if is user
            if($this->user_type == 1){
                $doctor = User::find()->where( ['id'=>$options['doctor_id']])->one();
                $message = [
                    "message" =>  [
                                    'notification_type' => Notification::TYPE_QUESTION,
                                    "theme_id"          => $options['object_id'],
                                    "user_id"           => $doctor->id,
                                    "user_name"         => $this->first_name.' '. $this->last_name,
                                    "image"             => $url.$doctor->image,
                                    "specialti"         => $doctor->getSpecialty()->name,
                                    "usertype"          => $this->user_type,
                                    "message"           => $text
                                  ]
                ];
            }

            // if is doctor
            if($this->user_type == 2){
                $user = User::find()->where(['id'=>$options['user_id']])->one();
                $message = [
                    "message" => [
                                    'notification_type' => Notification::TYPE_QUESTION,
                                    "theme_id"          => $options['object_id'],
                                    "user_id"           => $user->id,
                                    "user_name"         => $this->first_name.' '. $this->last_name,
                                    "image"             => $url.$user->image,
                                    "specialti"         => "null",
                                    "usertype"          => $user->user_type,
                                    "message"           => $text
                                 ]
                ];
            }

        }

        // if notification type == medHistory
        if($options['type'] == Notification::TYPE_MED_HISTORY){
            $message = [
                "message" => [
                    'notification_type'  => Notification::TYPE_MED_HISTORY,
                    "object_id"          => $options['object_id'],
                    "user_id"            => $this->id,
                    "text"               => $text,
                ]
            ];
        }

    }

    public function sendNotification($text,$options,$sms=true,$push=true){

        $notification = new Notification();

        $notification->sendNotification($text,$this->id,$options['type'],$options['message_type'],$options['object_id']);

        //if($sms)
          // $this->sendSms($text);

        if($push){
         //   $message = $this->getPushMessage($text,$options);
       //     $this->sendPushNotification($message);
        }
    }

    public function getLastLogin() {
            if($this->last_login != '0000-00-00 00:00:00')
                return $this->relativeTime(strtotime($this->last_login));
    }

    public function isOnline() {
        if($this->last_login != '0000-00-00 00:00:00'){
            $time = strtotime($this->last_login);
            if($time > strtotime('-3 minutes'))
                return true;
        }
        return false;
    }

    public function relativeTime($time, $short = false) {
        
        $SECOND = 1;
        $MINUTE = 60 * $SECOND;
        $HOUR   = 60 * $MINUTE;
        $DAY    = 24 * $HOUR;
        $MONTH  = 30 * $DAY;
        $before = time() - $time;
    
        if ($before < 2 * $MINUTE)
        {
           return '<span>'.Yii::t('app','on-line').'</span>';
        }else
        if ($before < 59 * $MINUTE)
        {
           return Yii::t('app',"{minute} minutes ago",['minute'=>floor($before / 60)]);
        
        }else if ($before < 90 * $MINUTE)
        {
            return Yii::t('app',"an hour ago");
        }else if ($before < 24 * $HOUR)
        {   
            if(floor($before / 60 / 60) > 4)
                return Yii::t('app',"{hour} hours ago",['hour'=>floor($before / 60 / 60)]);
            else
                return Yii::t('app',"{hour} hour ago",['hour'=>floor($before / 60 / 60)]);
        }else
            return Yii::t('app',"last login {date}",['date'=>date('d.m.Y',$time)]);
           
        return false;
    }

    public function isRole($roleName) {

        $assignments = Yii::$app->authManager->getAssignments($this->id);

        if(array_key_exists($roleName,$assignments))
            return true;
        else
            return false;
    }

    public function deleteImage()
    {
        $fileDir = Yii::getAlias('@webroot/'.$this->oldAttributes['image']);
        if(file_exists($fileDir) AND trim($this->oldAttributes['image']) != '') {
            unlink($fileDir);
            return true;
        }else
            return false;
    }

    public function getReviewsCount(){

        $cacheKey = "user_review_count_".$this->className().'_'.$this->id;

        $count = Yii::$app->cache->get($cacheKey);

        if ($count === false) {
            $count = Review::find()->where(['object_id'=>$this->id,'object_class_name'=>$this->className()])->count();
            Yii::$app->cache->set(
                $cacheKey,
                $count,
                86400,
                new TagDependency([
                    'tags' => $this->className().'_'.$this->id
                ])
            );
        }
        return $count;
    }

    public function getRaringCount(){

        $cacheKey = "user_rating_count_".$this->className().'_'.$this->id;
        $sum = Yii::$app->cache->get($cacheKey);
        if ($sum === false) {
            $sum = Rating::find()->where(['object_id'=>$this->id,'object_class_name'=>$this->className()])->sum('value');
            $count = Rating::find()->where(['object_id'=>$this->id,'object_class_name'=>$this->className()])->count();
            $sum =  number_format(($sum / $count), 2, '.', '');
            $user           = User::find()->where(['id'=>$this->id])->one();
            $user->scenario = 'rating';
            $user->rating   = $sum;
            $user->save();

            $user->rating = $sum;

            Yii::$app->cache->set(
                $cacheKey,
                $sum,
                86400,
                new TagDependency([
                    'tags' => $this->className().'_'.$this->id
                ])
            );
        }
        return $sum;
    }

}
